import { reactive, shallowRef } from "vue"
import { reactiveComputed } from "@vueuse/core"

export interface Act {
    name: string
    scenes: Scene[]
}

export interface Scene {
    name: string
    steps: Step[]
}

export interface Step {
    position: ShowPosition
    cue: StepCue
    actorEntrances: string[]
    actorExits: string[]
    actorsOnStage: string[]
    props: PropMap
    hasChangedProps: boolean
    leftSpotTarget: string | null
    rightSpotTarget: string | null
    curtainState: "open" | "closed"
}

export type StepCue = {
    type: "TEXT",
    speaker: string
    text: string
    clarification?: string
} | {
    type: "MUSIC_START",
    title: string
    duration: number
} | {
    type: "MUSIC_END"
} | {
    type: "CURTAIN",
    state: "open" | "closed"
    whileMoving: boolean
} | {
    type: "LIGHTS"
    state: "on" | "off"
    whileFading: boolean
} | {
    type: "CUSTOM"
    text: string
}

export type PropMap = Record<PropPosition, string | null>
export type PropPosition =
    | "PROSCENIUM_LEFT"
    | "PROSCENIUM_CENTER"
    | "PROSCENIUM_RIGHT"
    | "CENTER"
    | "LEFT"
    | "RIGHT"
    | "BACKDROP"

export interface ShowState {
    position: ShowPosition
    message: string
    activeMusic: ShowMusic | null
    musicStartTime: number
    isLightBehindCurtainOn: boolean
}

export interface ShowMusic {
    title: string
    duration: number
}

export interface ShowPosition {
    act: number
    scene: number
    step: number
}

export interface Show {
    acts: Act[]
}

export const START_STEP: Step = {
    position: { act: -1, scene: 0, step: 0 },
    actorsOnStage: [],
    cue: {
        type: "CUSTOM",
        text: "Start"
    },
    props: {
        BACKDROP: null,
        LEFT: null,
        CENTER: null,
        RIGHT: null,
        PROSCENIUM_LEFT: null,
        PROSCENIUM_CENTER: null,
        PROSCENIUM_RIGHT: null
    },
    hasChangedProps: false,
    actorEntrances: [],
    actorExits: [],
    leftSpotTarget: null,
    rightSpotTarget: null,
    curtainState: "closed"
}

const START_SCENE: Scene = {
    name: "Start",
    steps: [START_STEP]
}

export const show = shallowRef<Show>({
    acts: []
})

export const state = reactive<ShowState>({
    position: START_STEP.position,
    message: "",
    activeMusic: null,
    musicStartTime: 0,
    isLightBehindCurtainOn: false
})

export function getStep(position: ShowPosition) {
    if (position.act === -1) return START_STEP
    return getScene(position).steps[position.step]
}

export function getScene(position: ShowPosition) {
    if (position.act === -1) return START_SCENE
    return getAct(position)!.scenes[position.scene]
}

export function getAct(position: ShowPosition) {
    if (position.act === -1) return null
    return show.value.acts[position.act]
}

export function getSceneIndex(position: ShowPosition) {
    let index = position.scene

    for (let actIndex = 0; actIndex < position.act; actIndex++) {
        index += show.value.acts[actIndex]!.scenes.length
    }

    return index
}

export function getActiveMusicAt(position: ShowPosition): ShowMusic | null {
    let activeMusic: ShowMusic | null = null

    for (let actIndex = 0; actIndex < show.value.acts.length; actIndex++) {
        const scenes = show.value.acts[actIndex].scenes

        for (let sceneIndex = 0; sceneIndex < scenes.length; sceneIndex++) {
            const scene = scenes[sceneIndex]

            for (let stepIndex = 0; stepIndex < scene.steps.length; stepIndex++) {
                const step = scene.steps[stepIndex]

                // SONG
                if (step.cue.type === "MUSIC_START") {
                    activeMusic = {
                        title: step.cue.title,
                        duration: step.cue.duration
                    }
                } else if (step.cue.type === "MUSIC_END") {
                    activeMusic = null
                }

                if (sceneIndex == position.scene && stepIndex == position.step) return activeMusic
            }
        }
    }

    return null
}

export const current = reactiveComputed<{
    act: Act | null
    scene: Scene
    sceneIndex: number
    step: Step
    activeMusic: ShowMusic | null
    allScenes: Scene[]
}>(() => ({
    act: getAct(state.position),
    scene: getScene(state.position),
    sceneIndex: getSceneIndex(state.position),
    step: getStep(state.position),
    activeMusic: state.activeMusic,
    allScenes: show.value.acts.flatMap(act => act.scenes)
}))

export function parseStringWithDetails(string: string) {
    const parts = string.split(" / ")
    if (parts.length === 1) return { main: parts[0], details: "" }
    return { main: parts[0], details: parts.slice(1).join(" / ") }
}

export function getNextValidPosition(start: ShowPosition): ShowPosition | null {
    if (start.act === START_STEP.position.act) return { act: 0, scene: 0, step: 0 }
    const acts = show.value.acts
    let { act, scene, step } = start

    step++

    if (step >= acts[start.act].scenes[start.scene].steps.length) {
        step = 0
        scene++

        if (scene >= acts[start.act].scenes.length) {
            scene = 0
            act++

            if (act >= acts.length) {
                return null
            }
        }
    }

    return { act, scene, step }
}

export function getPreviousValidPosition(start: ShowPosition): ShowPosition | null {
    let { act, scene, step } = start
    step--

    if (step < 0) {
        scene--

        if (scene < 0) {
            act--

            if (act < 0) {
                return START_STEP.position
            }

            scene = show.value.acts[act].scenes.length - 1
        }

        step = show.value.acts[act].scenes[scene].steps.length - 1
    }

    return { act, scene, step }
}
